#open "types";;
(* Pour que l'on ne puisse pas prévoir la suite des nombres aléatoires utilisés... *)
if vect_length sys__command_line = 2 then
  random__init (int_of_string sys__command_line.(1));;

(* validité d'un employe *)
type validite =
  | Travaillant
  | Triche (* a tenté de travailler plus de 1000 heures dans l'annee *)
  | Epuise (* a travaillé 1000h et n'a pas gagné de prime *)
  | Licencie (* licencié : prime trop négative *)
  | Gagnant (* prime payée, il a gagné cette partie *);;

type comparaison = Gt (* Greater Than *)| Eq (* Equal *) | Lt (* Lower Than *);;
let key_maxima compare list =
  let (hd_key, hd_elt) = hd list in
  let extremum = ref hd_elt in
  let result = ref [hd_key] in
  do_list
    (fun (key,elt) ->
       match compare elt !extremum with
         | Gt -> result:=[key]; extremum:=elt
         | Eq -> result := key :: !result
         | Lt -> ())
    (tl list);
  !result;;

let une_partie numero_partie employes =
(*   random__init numero_partie; *)
  let nb_employes = vect_length employes in
  let validites = make_vect nb_employes Travaillant in
  let employes_valides = ref [] in
  for e=0 to nb_employes-1 do
    employes_valides := (e,{heures=nb_heures_init; nuts=0}) :: !employes_valides
  done;
  do_list (fun (e,etat) -> employes.(e).nouvelle_partie e nb_employes numero_partie) !employes_valides;
  (try while !employes_valides <> [] do
     let importance = 1+ random__int borne_importance in
(* printf__printf "importance:%d\n" importance; *)
     (* coups joués: *)
     let heures_travaillees = map
       (fun (e,etat) -> employes.(e).joue importance !employes_valides)
       !employes_valides in
     (* On décompte les heures travaillées (et on vire les tricheurs): *)
     let heures_travaillees, new_employes_valides = it_list2
       (fun (accu_h,accu_e) h (e,etat) ->
(* printf__printf "empl %d, etat: h=%d n=%d, joue:%d\n" *)
(* e etat.heures etat.nuts h; *)
          if h <= 0 || h > etat.heures
          then (validites.(e) <- Triche; accu_h,accu_e)
          else (e,h)::accu_h,  (e, {heures=etat.heures-h; nuts=etat.nuts}) :: accu_e)
       ([],[]) heures_travaillees !employes_valides in
     (* On informe les employés valides de ce qui vient d'être joué sans tricher: *)
     if new_employes_valides = [] then raise Exit;
     employes_valides := new_employes_valides;
     do_list (fun (e,etat) -> employes.(e).acquiesce heures_travaillees) !employes_valides;

     (* On cherche celui qui a travaillé le plus: *)
     (match
        key_maxima
          (fun h1 h2 -> if h1>h2 then Gt else if h1=h2 then Eq else Lt)
          heures_travaillees
      with
        | [employe_modele] ->
            employes_valides := it_list
              (fun accu ((e,etat) as employe) ->
                 if e = employe_modele
                 then (e, {nuts=etat.nuts+importance; heures=etat.heures}) :: accu
                 else employe :: accu)
              [] !employes_valides;
        | _::_::_ -> ()
        | [] -> failwith "key_maxima a renvoyé une liste vide");

     (* On cherche ceux qui ont travaillé le moins: *)
     let flemmards = key_maxima
       (fun h1 h2 -> if h1<h2 then Gt else if h1=h2 then Eq else Lt)
       heures_travaillees in
     employes_valides := it_list
       (fun accu ((e,etat) as employe) ->
          if mem e flemmards
          then (e,{nuts=etat.nuts-importance; heures=etat.heures}) :: accu
          else employe :: accu)
       [] !employes_valides;

     (* On cherche un éventuel gagnant: *)
     do_list (fun (e,etat) ->
          if etat.nuts >= prime_pr_gagner
          then (validites.(e) <- Gagnant; raise Exit))
       !employes_valides;
     (* On cherche les employés licenciés: *)
     employes_valides := it_list
       (fun accu ((e,etat) as employe) ->
          if etat.nuts <= - prime_pr_gagner
          then (validites.(e) <- Licencie; accu)
          else employe :: accu)
       [] !employes_valides;
     (* On cherche les employés épuisés: *)
     employes_valides := it_list
       (fun accu ((e,etat) as employe) ->
          if etat.heures = 0
          then (validites.(e) <- Epuise; accu)
          else employe :: accu)
       [] !employes_valides;
   done;
   with Exit -> ());
  validites;;

let int_of_validite = function
  | Gagnant -> 0
  | Travaillant -> 1
  | Triche -> 2
  | Epuise -> 3
  | Licencie -> 4
;;

let classement nb_parties employes =
  let nb_employes = vect_length employes in
  let stats = make_matrix nb_employes 5 0 in
  for i=0 to nb_parties-1 do
    let resultat = une_partie i employes in
    for i=0 to nb_employes-1 do
      let v = int_of_validite resultat.(i) in
      stats.(i).(v) <- stats.(i).(v) +1
    done
  done;
  stats;;

let swap vect i j =
  let temp = vect.(i) in
  vect.(i) <- vect.(j);
  vect.(j) <- temp;;

let shuffle vect =
  for i=1 to vect_length vect -1 do
    swap vect (random__int (i+1)) i
  done;;

let affiche nb_parties employes noms =
  let nb_employes = vect_length employes in
  (* On mélange les employés: *)
  let melange = init_vect nb_employes (fun i -> i) in
  shuffle melange;
  let employes = init_vect nb_employes (fun i -> employes.(melange.(i))) in
  let noms     = init_vect nb_employes (fun i ->     noms.(melange.(i))) in
  (* On calcule l'ordre des employés, du meilleur au moins bon: *)
  let stats = classement nb_parties employes in
  let ordre = init_vect nb_employes (fun i -> i) in
  let ordre = list_of_vect ordre in
  let ordre = sort__sort (fun e1 e2 ->
      (stats.(e1).(0), -stats.(e1).(2), -stats.(e1).(4))
    > (stats.(e2).(0), -stats.(e2).(2), -stats.(e2).(4))) ordre
  in
  let ordre = vect_of_list ordre in
  (* On affiche: *)
  printf__printf " %26s | gagnant | travaillant | triche | épuisé | licencié\n" "nom";
  for e=0 to nb_employes -1 do
    let stat = stats.(ordre.(e)) in
    printf__printf " %26s | %7d | %11d | %6d | %6d | %8d\n"
      noms.(ordre.(e)) stat.(0) stat.(1) stat.(2) stat.(3) stat.(4);
  done;
  flush stdout;;
